import uvicorn
from typer import Option, Typer
from typing_extensions import Annotated

from deciphon_sched.settings import Settings

app = Typer()

RELOAD = Annotated[bool, Option(help="Enable auto-reload.")]

@app.command()
def main(reload: RELOAD = False):
    settings = Settings()
    uvicorn.run(
        "deciphon_sched.main:app",
        host=settings.host,
        port=settings.port,
        reload=reload,
        log_level=settings.log_level.value,
    )



from fastapi import HTTPException
from sqlalchemy.exc import IntegrityError
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.status import (
    HTTP_400_BAD_REQUEST,
    HTTP_404_NOT_FOUND,
    HTTP_422_UNPROCESSABLE_ENTITY,
)


async def integrity_error_handler(_: Request, exc: IntegrityError):
    return JSONResponse(
        content={"detail": str(exc)}, status_code=HTTP_422_UNPROCESSABLE_ENTITY
    )


class FileNameExistsError(HTTPException):
    def __init__(self, name: str):
        super().__init__(
            HTTP_422_UNPROCESSABLE_ENTITY, f"File name '{name}' already exists"
        )


class FileNameNotFoundError(HTTPException):
    def __init__(self, name: str):
        super().__init__(HTTP_422_UNPROCESSABLE_ENTITY, f"File name '{name}' not found")


class NotFoundInDatabaseError(HTTPException):
    def __init__(self, name: str):
        super().__init__(HTTP_404_NOT_FOUND, f"'{name}' not found in the database")


class FoundInDatabaseError(HTTPException):
    def __init__(self, name: str):
        super().__init__(
            HTTP_422_UNPROCESSABLE_ENTITY, f"'{name}' already exists in the database"
        )


class JobStateTransitionError(HTTPException):
    def __init__(self, previous: str, next: str):
        super().__init__(HTTP_400_BAD_REQUEST, f"{previous}->{next} is invalid")


class SnapFileError(HTTPException):
    def __init__(self, msg: str):
        super().__init__(HTTP_400_BAD_REQUEST, f"Snap file is invalid: {msg}")



from fastapi import APIRouter, Request
from starlette.status import HTTP_200_OK

from . import dbs, hmms, jobs, scans, seqs

router = APIRouter()

router.include_router(hmms.router)
router.include_router(dbs.router)
router.include_router(jobs.router)
router.include_router(seqs.router)
router.include_router(scans.router)


@router.get("/", summary="list of all endpoints", status_code=HTTP_200_OK)
def root(request: Request):
    routes = request.app.routes
    return [{"path": route.path, "name": route.name} for route in routes]



from typing import Annotated

from deciphon_core.schema import Gencode
from fastapi import APIRouter, Request
from starlette.status import HTTP_200_OK, HTTP_201_CREATED, HTTP_204_NO_CONTENT

from deciphon_sched.database import Database
from deciphon_sched.errors import (
    FileNameExistsError,
    FileNameNotFoundError,
    NotFoundInDatabaseError,
)
from deciphon_sched.journal import Journal
from deciphon_sched.sched.models import HMM
from deciphon_sched.sched.schemas import HMMFile, HMMRead, PressRequest
from deciphon_sched.storage import PresignedDownload, PresignedUpload, Storage

router = APIRouter()


@router.get("/hmms", status_code=HTTP_200_OK)
async def read_hmms(request: Request) -> list[HMMRead]:
    database: Database = request.app.state.database
    with database.create_session() as session:
        return [x.read_model() for x in HMM.get_all(session)]


@router.get("/hmms/presigned-upload/{name}", status_code=HTTP_200_OK)
async def presigned_hmm_upload(
    request: Request,
    name: Annotated[str, HMMFile.path_type],
) -> PresignedUpload:
    storage: Storage = request.app.state.storage
    database: Database = request.app.state.database

    with database.create_session() as session:
        x = HMM.get_by_name(session, name)
        if x is not None:
            raise FileNameExistsError(name)
        return storage.presigned_upload(name)


@router.get("/hmms/presigned-download/{name}", status_code=HTTP_200_OK)
async def presigned_hmm_download(
    request: Request,
    name: Annotated[str, HMMFile.path_type],
) -> PresignedDownload:
    storage: Storage = request.app.state.storage
    database: Database = request.app.state.database

    with database.create_session() as session:
        x = HMM.get_by_name(session, name)
        if x is None:
            raise FileNameNotFoundError(name)
        return storage.presigned_download(name)


@router.post("/hmms/", status_code=HTTP_201_CREATED)
async def create_hmm(
    request: Request, hmm: HMMFile, gencode: Gencode, epsilon: float
) -> HMMRead:
    storage: Storage = request.app.state.storage
    database: Database = request.app.state.database

    with database.create_session() as session:
        x = HMM.get_by_name(session, hmm.name)
        if x is not None:
            raise FileNameExistsError(hmm.name)

        if not storage.has_file(hmm.name):
            raise FileNameNotFoundError(hmm.name)

        x = HMM.create(hmm)
        session.add(x)
        session.commit()
        hmm_read = x.read_model()

    journal: Journal = request.app.state.journal
    x = PressRequest.create(hmm_read.job.id, hmm_read.file, gencode, epsilon)
    await journal.publish("press", x.model_dump_json())

    return hmm_read


@router.get("/hmms/{hmm_id}", status_code=HTTP_200_OK)
async def read_hmm(request: Request, hmm_id: int) -> HMMRead:
    database: Database = request.app.state.database
    with database.create_session() as session:
        x = HMM.get_by_id(session, hmm_id)
        if x is None:
            raise NotFoundInDatabaseError("HMM")
        return x.read_model()


@router.delete("/hmms/{hmm_id}", status_code=HTTP_204_NO_CONTENT)
async def delete_hmm(request: Request, hmm_id: int):
    storage: Storage = request.app.state.storage
    database: Database = request.app.state.database

    with database.create_session() as session:
        x = HMM.get_by_id(session, hmm_id)
        if x is None:
            raise NotFoundInDatabaseError("HMM")
        if storage.has_file(x.name):
            storage.delete(x.name)
        session.delete(x)
        session.commit()



from aiomqtt import Client
from deciphon_sched.logger import Logger

from deciphon_sched.settings import Settings


class Journal:
    def __init__(self, settings: Settings, logger: Logger):
        self._mqtt = Client(hostname=settings.mqtt_host, port=settings.mqtt_port)
        self._topic = settings.mqtt_topic
        self._logger = logger

    async def __aenter__(self):
        await self._mqtt.__aenter__()
        return self

    async def __aexit__(self, *args, **kargs):
        await self._mqtt.__aexit__(*args, **kargs)

    async def publish(self, subject: str, payload: str):
        topic = f"/{self._topic}/{subject}"
        self._logger.handler.info(f"publishing <{payload}> to <{topic}>")
        await self._mqtt.publish(topic, payload)



from pydantic import HttpUrl, AnyUrl
from pydantic_settings import BaseSettings, SettingsConfigDict
from enum import Enum


class LogLevel(str, Enum):
    debug = "debug"
    info = "info"
    warning = "warning"
    error = "error"
    critical = "critical"

    # Enum of Python3.10 returns a different string representation.
    # Make it return the same as in Python3.11
    def __str__(self):
        return str(self.value)


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="deciphon_sched_")

    host: str = "0.0.0.0"
    port: int = 8000

    endpoint_prefix: str = ""
    allow_origins: list[str] = ["http://127.0.0.1", "http://localhost"]
    log_level: LogLevel = LogLevel.info

    database_url: AnyUrl = AnyUrl("sqlite+pysqlite:///:memory:")

    s3_key: str = "minioadmin"
    s3_secret: str = "minioadmin"
    s3_url: HttpUrl = HttpUrl("http://localhost:9000")
    s3_bucket: str = "deciphon"

    mqtt_host: str = "localhost"
    mqtt_port: int = 1883
    mqtt_topic: str = "deciphon"
